#include "Realizer.h"

Realizer::Realizer( hub::Hub* theHub ) : theHub(theHub){
}

Realizer::~Realizer(){
}

void Realizer::operator() ( hub::SrcLocation& src, hub::DestLocation& dest, RealizationAlgorithmID algorithmID ){
	//copy ports
	for( hub::Entity::PortProtoIter it = src.theEntity()->PortProtoBeg(); it != src.theEntity()->PortProtoEnd(); it++ ){
		dest.theEntity()->CreateNextPort( it->first, it->second->flags() );
	}

	//prepare location
	hub::StructArch* destArchImpl = new hub::StructArch;
	dest.theArch()->Impl( destArchImpl );
	dest.theEntity()->CurrentArch( dest.theArch() );

	//instantiate ports
	for( hub::Entity::PortProtoIter pit = dest.theEntity()->PortProtoBeg(); pit != dest.theEntity()->PortProtoEnd(); ++pit ){
		destArchImpl->CreatePort( pit->second );
	}

	const hub::FsmDescr* srcFSMDescr = src.theArch()->Impl<hub::BehavArch>()->description<hub::FsmDescr>();

	switch( algorithmID ){
		///////////////////////////////////////////////////
		//add your new algorithm implementation call here//
		///////////////////////////////////////////////////
		case naiveRTLDFFbased: this->doNaiveRTLDFFbased( srcFSMDescr, destArchImpl ); break;
	}
}


bool Realizer::isBinary( std::string str ){
	for( std::string::iterator it = str.begin(); it < str.end(); it++ ){
		if( *it != '1' && *it != '0' && *it != '-' ) return false;
	}
	return true;
}

std::string Realizer::i2Str( unsigned ui ){
	std::stringstream st;
	st << ui;
	return st.str();
}

std::string Realizer::i2Str( int i ){
	std::stringstream st;
	st << i;
	return st.str();
}

std::list< hub::StateTableRow >& Realizer::createTransitionTable( const hub::FsmDescr* fsm ){
	std::list<hub::StateTableRow>* stateTable = new std::list<hub::StateTableRow>;
	fsm->CreateTable( *stateTable );

	for( std::list<hub::StateTableRow>::iterator it = stateTable->begin(); it != stateTable->end(); it++ ){

		if( !( isBinary( it->input ) && isBinary( it->output ) ) ){
			throw hub::CException< std::runtime_error, hub::InconsReason, hub::ToolPosition >(
				hub::InconsReason( "fsm inputs or outputs are not binary" ),
				hub::ToolPosition( "Realizer::createTransitionTable()" )
			);
		}

		it->currState = fsm->FindState( it->currState )->encoding();
		it->nextState = fsm->FindState( it->nextState )->encoding();
		
		if( !( isBinary( it->currState ) && isBinary( it->nextState ) && it->currState.length() != 0 && it->nextState.length() != 0 ) ){
			throw hub::CException< std::runtime_error, hub::InconsReason, hub::ToolPosition >(
				hub::InconsReason( "fsm states are not encoded (/in binary)" ),
				hub::ToolPosition( "Realizer::createTransitionTable()" )
			);
		}
	}
	
	return *stateTable;
}

hub::Entity* Realizer::createFuncEntity( std::string name, hub::SymFunc::Func func, unsigned inputCount, hub::Library* destLibrary ){
	if( name.length() < 1 || destLibrary==0 ){
		throw hub::CException< std::runtime_error, hub::InconsReason, hub::ToolPosition >(
				hub::InconsReason( "createFuncEntity() arguments are invalid" ),
				hub::ToolPosition( "Realizer::createFuncEntity()" )
			);
	}

	hub::Entity* funcEntity = destLibrary->FindOrCreateEntity( name );
	hub::Architecture* funcArchitecture = funcEntity->CreateArch<hub::BehavArch>( name + "architecture" );
	funcEntity->CurrentArch( funcArchitecture );
	hub::BehavArch* funcArchImpl = new hub::BehavArch;
	funcArchitecture->Impl( funcArchImpl );

	hub::SymFunc* funcDescr = new hub::SymFunc;
	funcDescr->func( func );
	funcArchImpl->description( funcDescr );

	funcDescr->inpNo( inputCount );

	funcEntity->CreateNextPort( "OUT", hub::PortProto::outDir );
	for( unsigned i = 0; i < inputCount; ++i){
		funcEntity->CreateNextPort( "IN[" + i2Str( i ) + "]", hub::PortProto::inDir );
	}

	return funcEntity;
}

hub::Instance* Realizer::createStateMemory( std::string name, hub::BehavAgent* memoryElement, unsigned inputCount, hub::Library* destLibrary, hub::StructArch* destArch ){
	//memory element
	hub::Entity* dffEntity = destLibrary->CreateEntity( name );
	hub::Architecture* dffArchitecture = dffEntity->CreateArch<hub::BehavArch>( name + "architecture" );
	hub::BehavArch* dffArchImpl = new hub::BehavArch;
	dffArchitecture->Impl( dffArchImpl );
	dffArchImpl->description( memoryElement->CreateDescr() );
	dffEntity->CreateNextPort( "D", hub::PortProto::dataInDir );
	dffEntity->CreateNextPort( "Q", hub::PortProto::outDir );
	dffEntity->CreateNextPort( "CLK", hub::PortProto::clkInDir );
	dffEntity->CreateNextPort( "SET", hub::PortProto::inDir );
	dffEntity->CreateNextPort( "RST", hub::PortProto::inDir );

	//D FF *n state memory - with inputsCount=encodingDigits
	hub::Entity* memoryEntity = destLibrary->CreateEntity( name + i2Str( inputCount ) );
	hub::Architecture* memoryArchitecture = memoryEntity->CreateArch<hub::StructArch>( name + i2Str( inputCount ) + "architecture" );
	memoryEntity->CurrentArch( memoryArchitecture );
	hub::StructArch* memoryArchImpl = new hub::StructArch;
	memoryArchitecture->Impl( memoryArchImpl );

	memoryEntity->CreateNextPort( "CLK", hub::PortProto::clkInDir );
	for( unsigned i=0; i < inputCount; ++i ){
		memoryEntity->CreateNextPort( "D[" + i2Str( i ) + "]", hub::PortProto::dataInDir );
		memoryEntity->CreateNextPort( "Q[" + i2Str( i ) + "]", hub::PortProto::outDir );
		memoryEntity->CreateNextPort( "SET[" + i2Str( i ) + "]", hub::PortProto::inDir );
		memoryEntity->CreateNextPort( "RST[" + i2Str( i ) + "]", hub::PortProto::inDir );
	}

	for( hub::Entity::PortProtoIter pit = memoryEntity->PortProtoBeg(); pit != memoryEntity->PortProtoEnd(); ++pit ){
		memoryArchImpl->CreatePort( pit->second );
	}


	hub::Net* clkNet = memoryArchImpl->CreateNet( "CLK" );
	clkNet->Join( memoryArchImpl->FindPort( "CLK" ) );

	for( unsigned i=0; i < inputCount; ++i){
		hub::Instance* dffInstance = dffEntity->Instantiate( name + "[" + i2Str( i ) + "]", memoryArchImpl, dffEntity->CurrentArch() );

		clkNet->Join( dffInstance->FindPort( "CLK" ) );

		hub::Net* net = memoryArchImpl->CreateNet( "D" + i2Str( i ) );
		net->Join( memoryArchImpl->FindPort( "D[" + i2Str( i ) + "]" ) );
		net->Join( dffInstance->FindPort( "D" ) );

		net = memoryArchImpl->CreateNet( "Q" + i2Str( i ) );
		net->Join( memoryArchImpl->FindPort( "Q[" + i2Str( i ) + "]" ) );
		net->Join( dffInstance->FindPort( "Q" ) );

		net = memoryArchImpl->CreateNet( "SET" + i2Str( i ) );
		net->Join( memoryArchImpl->FindPort( "SET[" + i2Str( i ) + "]" ) );
		net->Join( dffInstance->FindPort( "SET" ) );

		net = memoryArchImpl->CreateNet( "RST" + i2Str( i ) );
		net->Join( memoryArchImpl->FindPort( "RST[" + i2Str( i ) + "]" ) );
		net->Join( dffInstance->FindPort( "RST" ) );
	}
	
	hub::Instance* memoryInstance = memoryEntity->Instantiate( name + "stateMemory", destArch, memoryEntity->CurrentArch() );
	return memoryInstance;
}

hub::Net* Realizer::createMintermNet(	unsigned j, unsigned stateCount, unsigned inputCount, std::list< hub::StateTableRow >::iterator it, hub::StructArch* decoderArchImpl,
										hub::Entity* andEntity, hub::Entity* invEntity, hub::Instance* vccInstance ){
	//create minterm
	hub::Instance* mintermInstance = andEntity->Instantiate( "minterm" + i2Str( j ), decoderArchImpl, andEntity->CurrentArch() );
	hub::Net* mintermNet = decoderArchImpl->CreateNet( "minterm" + i2Str( j ) );
	mintermNet->Join( mintermInstance->FindPort( "OUT" ) );

	//for every input binary in this line
	//for every currentState input
	for( unsigned i = 0; i < stateCount; ++i ){
		//connect to AND mintermInstance
		hub::Net* andNet = decoderArchImpl->CreateNet( "andstate" + i2Str( i ) + "-" + i2Str( j ) );
		andNet->Join( mintermInstance->FindPort( "IN[" + i2Str( i ) + "]" ) );

		//connect 1es directly and 0es inverted, DC as VCC
		if( it->currState[i] == '1' ){
			andNet->Join( decoderArchImpl->FindPort( "INSTATE[" + i2Str( i ) + "]" ) );
		}
		else if( it->currState[i] == '0' ){
			hub::Instance* invInstance = invEntity->Instantiate( "invinstate" + i2Str( i ) + "-" + i2Str( j ), decoderArchImpl, invEntity->CurrentArch() );
			hub::Net* invNet = decoderArchImpl->CreateNet( "invstate" + i2Str( i ) + "-" + i2Str( j ) );
			invNet->Join( invInstance->FindPort( "IN[0]" ) );
			invNet->Join( decoderArchImpl->FindPort( "INSTATE[" + i2Str( i ) + "]" ) );
			andNet->Join( invInstance->FindPort( "OUT" ) );
		}
		else{
			andNet->Join( vccInstance->FindPort( "OUT" ) );
		}

	}

	//for every Input input
	for( unsigned i = 0; i < inputCount; ++i ){
		//connect to AND mintermInstance
		unsigned iplus = i + stateCount;

		hub::Net* andNet = decoderArchImpl->CreateNet( "andinput" + i2Str( i ) + "-" + i2Str( j ) );
		andNet->Join( mintermInstance->FindPort( "IN[" + i2Str( iplus ) + "]" ) );

		//connect 1es directly and 0es inverted, DC as VCC
		if( it->input[i] == '1' ){
			andNet->Join( decoderArchImpl->FindPort( "IN[" + i2Str( i ) + "]" ) );
		}
		else if( it->input[i] == '0' ){
			hub::Instance* invInstance = invEntity->Instantiate( "invininput" + i2Str( i ) + "-" + i2Str( j ), decoderArchImpl, invEntity->CurrentArch() );
			hub::Net* invNet = decoderArchImpl->CreateNet( "invinput" + i2Str( i ) + "-" + i2Str( j ) );
			invNet->Join( invInstance->FindPort( "IN[0]" ) );
			invNet->Join( decoderArchImpl->FindPort( "IN[" + i2Str( i ) + "]" ) );
			andNet->Join( invInstance->FindPort( "OUT" ) );
		}
		else{
			andNet->Join( vccInstance->FindPort( "OUT" ) );
		}

	}

	return mintermNet;
}

hub::Instance* Realizer::createTransitionsDecoder(	std::string name, std::list< hub::StateTableRow >& transitionsTable, hub::Library* destLibrary, hub::StructArch* destArch,
													hub::Entity* andEntity, hub::Entity* orEntity, hub::Entity* invEntity, hub::Entity* vccEntity, hub::Entity* gndEntity ){
	//create hierarchy
	unsigned inputCount = transitionsTable.begin()->input.length();
	unsigned outputCount = transitionsTable.begin()->output.length();
	unsigned stateCount = transitionsTable.begin()->currState.length();
	unsigned nextStateCount = transitionsTable.begin()->nextState.length();

	hub::Entity* decoderEntity = destLibrary->FindOrCreateEntity( name + i2Str( inputCount + stateCount ) + "to" + i2Str( outputCount + nextStateCount ) );
	hub::Architecture* decoderArchitecture = decoderEntity->CreateArch<hub::StructArch>( name + i2Str( inputCount + stateCount ) + "to" + i2Str( outputCount + nextStateCount ) + "architecture" );
	decoderEntity->CurrentArch( decoderArchitecture );
	hub::StructArch* decoderArchImpl = new hub::StructArch;
	decoderArchitecture->Impl( decoderArchImpl );
	
	//generate ports
	for( unsigned i=0; i < inputCount; ++i ){
		decoderEntity->CreateNextPort( "IN[" + i2Str( i ) + "]", hub::PortProto::inDir );
	}
	for( unsigned i=0; i < outputCount; ++i ){
		decoderEntity->CreateNextPort( "OUT[" + i2Str( i ) + "]", hub::PortProto::outDir );
	}
	for( unsigned i=0; i < stateCount; ++i ){
		decoderEntity->CreateNextPort( "INSTATE[" + i2Str( i ) + "]", hub::PortProto::inDir );
	}
	for( unsigned i=0; i < nextStateCount; ++i ){
		decoderEntity->CreateNextPort( "OUTSTATE[" + i2Str( i ) + "]", hub::PortProto::outDir );
	}
	
	for( hub::Entity::PortProtoIter pit = decoderEntity->PortProtoBeg(); pit != decoderEntity->PortProtoEnd(); ++pit ){
		decoderArchImpl->CreatePort( pit->second );
	}

	//connect output ORs
	std::vector<hub::Instance*> orStateInstances;
	std::vector<hub::Instance*> orOutputInstances;

	for( unsigned i=0; i < outputCount; ++i ){
		hub::Instance* orInstance = orEntity->Instantiate( "mintermSum" + i2Str( i ), decoderArchImpl, orEntity->CurrentArch() );
		hub::Net* outNet = decoderArchImpl->CreateNet( "output" + i2Str( i ) );
		outNet->Join( orInstance->FindPort( "OUT" ) );
		outNet->Join( decoderArchImpl->FindPort( "OUT[" + i2Str( i ) + "]" ) );

		orOutputInstances.push_back( orInstance );
	}
	for( unsigned i=0; i < nextStateCount; ++i ){
		unsigned iplus = i + outputCount;
		hub::Instance* orInstance = orEntity->Instantiate( "mintermSum" + i2Str( iplus ), decoderArchImpl, orEntity->CurrentArch() );
		hub::Net* outNet = decoderArchImpl->CreateNet( "outState" + i2Str( i ) );
		outNet->Join( orInstance->FindPort( "OUT" ) );
		outNet->Join( decoderArchImpl->FindPort( "OUTSTATE[" + i2Str( i ) + "]" ) );

		orStateInstances.push_back( orInstance );
	}

	//create GND and VCC instance
	hub::Instance* gndInstance = gndEntity->Instantiate( "decGND", decoderArchImpl, gndEntity->CurrentArch() );
	hub::Instance* vccInstance = vccEntity->Instantiate( "decVCC", decoderArchImpl, vccEntity->CurrentArch() );
	hub::Net* gndNet = decoderArchImpl->CreateNet( "GND" );
	//hub::Net* vccNet = decoderArchImpl->CreateNet( "VCC" );
	gndNet->Join( gndInstance->FindPort( "OUT" ) );
	//vccNet->Join( vccInstance->FindPort( "OUT" ) );

	//create sum of products canonical form as decoder
	//for every line in transition table create minterm
	//then connect minterm to OR of output if 1
	unsigned j = 0;
	//for every transition line
	for( std::list< hub::StateTableRow >::iterator it = transitionsTable.begin(); it != transitionsTable.end(); it++ ){
		hub::Net* mintermNet = 0;

		//for every output for 1 connect minterm else connect GND
		for( unsigned i = 0; i < nextStateCount; ++i ){
			if( it->nextState[i] == '1' ){
				if( mintermNet == 0 ){
					mintermNet = Realizer::createMintermNet( j, stateCount, inputCount, it, decoderArchImpl, andEntity, invEntity, vccInstance );
				}
				mintermNet->Join( orStateInstances[i]->FindPort( "IN[" + i2Str( j ) + "]" ) );
			}
			else{
				gndNet->Join( orStateInstances[i]->FindPort( "IN[" + i2Str( j ) + "]" ) );
			}
		}
		for( unsigned i = 0; i < outputCount; ++i ){
			if( it->output[i] == '1' ){
				if( mintermNet == 0 ){
					mintermNet = Realizer::createMintermNet( j, stateCount, inputCount, it, decoderArchImpl, andEntity, invEntity, vccInstance );
				}
				mintermNet->Join( orOutputInstances[i]->FindPort( "IN[" + i2Str( j ) + "]" ) );
			}
			else{
				gndNet->Join( orOutputInstances[i]->FindPort( "IN[" + i2Str( j ) + "]" ) );
			}
		}
		
		++j;
	}

	//clean redundant nets and ANDs
	/*
	for( hub::StructArch::NetIter nit = decoderArchImpl->NetBeg(); nit != decoderArchImpl->NetEnd(); ++nit ){
		hub::Net* net = nit->second;
		//if only one join
		if( ++( net->JoinBeg() ) == net->JoinEnd() ){
			hub::Instance* redInstance = dynamic_cast<hub::Instance*>( net->JoinBeg() )->Parent();

		}
	}
	*/
	
	hub::Instance* decoderInstance = decoderEntity->Instantiate( name + "transitionsDecoder", destArch, decoderEntity->CurrentArch() );
	return decoderInstance;
}

void Realizer::doNaiveRTLDFFbased( const hub::FsmDescr* srcFSMDescr, hub::StructArch* destArchImpl ){
	hub::Library* destLibrary = destArchImpl->Parent()->Parent()->Parent();
	//add clock and reset port next to the fsm ports
	hub::Entity* destEntity = destArchImpl->Parent()->Parent();
	hub::PortProto* clkPortProto = destEntity->CreateNextPort( "CLK", hub::PortProto::clkInDir );
	destArchImpl->CreatePort( clkPortProto );
	hub::PortProto* rstPortProto = destEntity->CreateNextPort( "RST", hub::PortProto::inDir );
	destArchImpl->CreatePort( rstPortProto );


	//hub::Instance* destInstance = destEntity->Instantiate( "", destArchImpl, destEntity->CurrentArch() );

	std::list< hub::StateTableRow > transitionTable = Realizer::createTransitionTable( srcFSMDescr );

	//prepare building blocks
	//////////////////////////////////////////////////

	//D FF *n state memory - with inputsCount=encodingDigits
	hub::Instance* stateMemoryInstance = Realizer::createStateMemory( "DFF", new hub::RTLRegAgent, srcFSMDescr->resetState()->encoding().length(), destLibrary, destArchImpl );

	//AND gate for transition function
	unsigned mintermMemberCount = srcFSMDescr->inputNo() + srcFSMDescr->resetState()->encoding().length();
	hub::Entity* andEntity = Realizer::createFuncEntity( "AND" + i2Str( mintermMemberCount ), hub::SymFunc::fAnd, mintermMemberCount, destLibrary );

	//INV gate
	hub::Entity* invEntity = Realizer::createFuncEntity( "INV", hub::SymFunc::fInv, 1, destLibrary );

	//OR gates for transition function
	unsigned maxMintermsCount = transitionTable.size();
	hub::Entity* orEntity = Realizer::createFuncEntity( "OR" + i2Str( maxMintermsCount ), hub::SymFunc::fOr, maxMintermsCount, destLibrary );

	// GND and VCC entity
	hub::Entity* gndEntity = Realizer::createFuncEntity( "GND", hub::SymFunc::fT0, 0, destLibrary );
	hub::Entity* vccEntity = Realizer::createFuncEntity( "VCC", hub::SymFunc::fT1, 0, destLibrary );

	//create sequential part
	//////////////////////////////////////////////////

	//connect state memory to clock
	hub::Net* clkNet = destArchImpl->CreateNet( "CLK" );
	clkNet->Join( destArchImpl->FindPort( "CLK" ) );
	clkNet->Join( stateMemoryInstance->FindPort( "CLK" ) );

	//create reset to initial state part
	//for 0 in encoded initial state connect global reset to reset and gnd to set
	//for 1 in encoded initial state connect global reset to set and gnd to reset
	hub::Net* rstNet = destArchImpl->CreateNet( "RST" );
	rstNet->Join( destArchImpl->FindPort( "RST" ) );
	hub::Instance* gndInstance = gndEntity->Instantiate( "GND", destArchImpl, gndEntity->CurrentArch() );
	hub::Net* gndNet = destArchImpl->CreateNet( "GND" );
	gndNet->Join( gndInstance->FindPort( "OUT" ) );

	std::string resetStateCode = srcFSMDescr->resetState()->encoding();
	for( unsigned i = 0; i < resetStateCode.length(); ++i ){
		if( resetStateCode[ i ] == '0' ){
			gndNet->Join( stateMemoryInstance->FindPort( "SET[" + i2Str( i ) + "]" ) );
			rstNet->Join( stateMemoryInstance->FindPort( "RST[" + i2Str( i ) + "]" ) );
		}
		else{
			gndNet->Join( stateMemoryInstance->FindPort( "RST[" + i2Str( i ) + "]" ) );
			rstNet->Join( stateMemoryInstance->FindPort( "SET[" + i2Str( i ) + "]" ) );
		}
	}

	//create combinational part
	//////////////////////////////////////////////////
	hub::Instance* decoderInstance = Realizer::createTransitionsDecoder( "decoder", transitionTable, destLibrary, destArchImpl, andEntity, orEntity, invEntity, vccEntity, gndEntity );
	
	unsigned i = 0;
	unsigned o = 0;

	for( hub::StructArch::PortIter pit = destArchImpl->PortBeg(); pit != destArchImpl->PortEnd(); pit++ ){

		hub::Port<hub::StructArch>* port = (*pit);
		if( port->predicate( hub::PortProto::input ) && port->name() != "CLK" && port->name() != "RST" ){
			hub::Net* net = destArchImpl->CreateNet( "IN" + i2Str(i) );
			net->Join( port );
			net->Join( decoderInstance->FindPort( "IN[" + i2Str(i) + "]" ) );

			++i;
		}
		else if( (*pit)->predicate( hub::PortProto::output ) ){
			hub::Net* net = destArchImpl->CreateNet( "OUT" + i2Str(o) );
			net->Join( port );
			net->Join( decoderInstance->FindPort( "OUT[" + i2Str(o) + "]" ) );

			++o;
		}
	}

	//seqential connects combinational
	//////////////////////////////////////////////////
	
	for( unsigned i = 0; i < srcFSMDescr->resetState()->encoding().length(); ++i ){
		hub::Net* inNet = destArchImpl->CreateNet( "input" + i2Str( i ) );
		inNet->Join( stateMemoryInstance->FindPort( "D[" + i2Str( i ) + "]" ) );
		inNet->Join( decoderInstance->FindPort( "OUTSTATE[" + i2Str( i ) + "]" ) );
		

		hub::Net* outNet = destArchImpl->CreateNet( "output" + i2Str( i ) );
		outNet->Join( stateMemoryInstance->FindPort( "Q[" + i2Str( i ) + "]" ) );
		outNet->Join( decoderInstance->FindPort( "INSTATE[" + i2Str( i ) + "]" ) );
	}

}
