#include "AlgorithmIdentificator.h"

//	AlgorithmIdentificator
//////////////////////////////////////////////////////////////////////////////////////////////
AlgorithmIdentificator::AlgorithmIdentificator( std::string name, std::string uniqueIdentificator )
	: name(name), uniqueIdentificator( uniqueIdentificator ){
}

AlgorithmIdentificator::~AlgorithmIdentificator(){
}


AlgorithmIdentificator::operator std::string() const{
	return this->uniqueIdentificator;
}

std::string AlgorithmIdentificator::getName() const{
	return this->name;
}

//	StateMinimizationAlgorithmIdentificator
//////////////////////////////////////////////////////////////////////////////////////////////
StateMinimizationAlgorithmIdentificator::StateMinimizationAlgorithmIdentificator( std::string name, std::string uniqueIdentificator, StateMinimizationAlgorithmID ID )
	: AlgorithmIdentificator( name, uniqueIdentificator ), ID( ID ){
}

StateMinimizationAlgorithmIdentificator::StateMinimizationAlgorithmIdentificator( std::string uniqueIdentificator ) : AlgorithmIdentificator( uniqueIdentificator, uniqueIdentificator ){
	int i = 0;
	for( ;; ){
		const StateMinimizationAlgorithmIdentificator algorithm = STATE_MINIMIZATION_ALGORITHMS_CHOICE[i];
		if( lastM == (StateMinimizationAlgorithmID)algorithm ){ 
			using namespace hub;
			throw CException <std::runtime_error, InconsReason, ToolPosition>( InconsReason ("Wrong state minimization algorithm - algorithm "+uniqueIdentificator+" does not exist"),
				ToolPosition ("Synthesizer::stateMinimizationAlgorithmIdentificator( std::string uniqueIdentificator )")
			);
		}
		else if( uniqueIdentificator == (std::string)algorithm ){
			*this = algorithm;
			break;
		}

		i++;
	}
}

StateMinimizationAlgorithmIdentificator::~StateMinimizationAlgorithmIdentificator(){
}

StateMinimizationAlgorithmIdentificator::operator StateMinimizationAlgorithmID() const{
	return this->ID;
}

//	EncodingAlgorithmIdentificator
//////////////////////////////////////////////////////////////////////////////////////////////

EncodingAlgorithmIdentificator::EncodingAlgorithmIdentificator( std::string name, std::string uniqueIdentificator, EncodingAlgorithmID ID )
	: AlgorithmIdentificator( name, uniqueIdentificator ), ID( ID ){
}

EncodingAlgorithmIdentificator::EncodingAlgorithmIdentificator( std::string uniqueIdentificator ) : AlgorithmIdentificator( uniqueIdentificator, uniqueIdentificator ){
	int i = 0;
	for( ;; ){
		const EncodingAlgorithmIdentificator algorithm = ENCODING_ALGORITHMS_CHOICE[i];
		if( lastE == algorithm ){
			using namespace hub;
			throw CException <std::runtime_error, InconsReason, ToolPosition>( InconsReason ("Wrong encoding algorithm - algorithm "+uniqueIdentificator+" does not exist"),
				ToolPosition ("Synthesizer::EncodingAlgorithmIdentificator( std::string uniqueIdentificator )")
			);
		}
		else if( uniqueIdentificator == (std::string)algorithm ){
			*this = algorithm;
			break;
		}

		++i;
	}
}

EncodingAlgorithmIdentificator::~EncodingAlgorithmIdentificator(){
}

EncodingAlgorithmIdentificator::operator EncodingAlgorithmID() const{
	return this->ID;
}

//	RealizationAlgorithmIdentificator
//////////////////////////////////////////////////////////////////////////////////////////////

RealizationAlgorithmIdentificator::RealizationAlgorithmIdentificator( std::string name, std::string uniqueIdentificator, RealizationAlgorithmID ID )
	: AlgorithmIdentificator( name, uniqueIdentificator ), ID( ID ){
}

RealizationAlgorithmIdentificator::RealizationAlgorithmIdentificator( std::string uniqueIdentificator ) : AlgorithmIdentificator( uniqueIdentificator, uniqueIdentificator ){
	int i = 0;
	for( ;; ){
		const RealizationAlgorithmIdentificator algorithm = REALIZATION_ALGORITHMS_CHOICE[i];
		if( lastR == algorithm ){
			using namespace hub;
			throw CException <std::runtime_error, InconsReason, ToolPosition>( InconsReason ("Wrong realization algorithm - algorithm "+uniqueIdentificator+" does not exist"),
				ToolPosition ("Synthesizer::RealizationAlgorithmIdentificator( std::string uniqueIdentificator )")
			);
		}
		else if( uniqueIdentificator == (std::string)algorithm ){
			*this = algorithm;
			break;
		}

		++i;
	}
}

RealizationAlgorithmIdentificator::~RealizationAlgorithmIdentificator(){
}

RealizationAlgorithmIdentificator::operator RealizationAlgorithmID() const{
	return this->ID;
}
